// Push and Pull support for Datman

use anyhow::{bail, ensure, Context};
use std::collections::{BTreeMap, BTreeSet};
use std::io::{Read, Write};
use std::sync::Arc;
use yama::chunking::RecursiveUnchunker;
use yama::commands::retrieve_tree_node;
use yama::definitions::{ChunkId, PointerData, RecursiveChunkRef, TreeNode};
use yama::pile::{Keyspace, Pile, PipelineDescription, RawPile};
use yama::progress::ProgressTracker;
use yama::remote::{read_message, write_message};

pub fn offer_pointers<W: Write, RP: RawPile>(
    pile: &Pile<RP>,
    writer: &mut W,
) -> anyhow::Result<BTreeMap<String, PointerData>> {
    let mut pointers_to_offer: BTreeMap<String, PointerData> = BTreeMap::new();

    for pointer_name in pile.list_pointers()? {
        let pointer_data = pile
            .read_pointer(&pointer_name)?
            .context("Listed pointer not present")?;

        pointers_to_offer.insert(pointer_name, pointer_data);
    }

    write_message(writer, &pointers_to_offer)?;
    Ok(pointers_to_offer)
}

pub fn ensure_compatible_bypasses(
    my_full: &Vec<PipelineDescription>,
    my_bypass: &Vec<PipelineDescription>,
    their_full: &Vec<PipelineDescription>,
    their_bypass: &Vec<PipelineDescription>,
) -> anyhow::Result<()> {
    ensure!(
        my_full.starts_with(&my_bypass),
        "Our full pipeline is not an extension of the bypass pipeline."
    );
    ensure!(
        their_full.starts_with(&their_bypass),
        "Their full pipeline is not an extension of their bypass pipeline."
    );

    let my_bypassed_parts = &my_full[my_bypass.len()..];
    let their_bypassed_parts = &their_full[their_bypass.len()..];

    ensure!(
        my_bypassed_parts == their_bypassed_parts,
        "Our bypassed parts and their bypassed parts are not the same.\nOurs: {:?}\nTheirs: {:?}",
        my_bypassed_parts,
        their_bypassed_parts
    );

    Ok(())
}

pub fn negotiate_bypassed_pile<R: Read, W: Write>(
    pile: &Pile<Arc<Box<dyn RawPile>>>,
    bypass_pile: &Box<dyn RawPile>,
    reader: &mut R,
    writer: &mut W,
) -> anyhow::Result<()> {
    let my_full_pipeline = pile.raw_pile.describe_pipeline()?;
    let my_bypass_pipeline = bypass_pile.describe_pipeline()?;

    write_message(writer, &my_full_pipeline)?;
    write_message(writer, &my_bypass_pipeline)?;
    writer.flush()?;

    let their_full_pipeline = read_message::<_, Vec<PipelineDescription>>(reader)?;
    let their_bypass_pipeline = read_message::<_, Vec<PipelineDescription>>(reader)?;

    ensure_compatible_bypasses(
        &my_full_pipeline,
        &my_bypass_pipeline,
        &their_full_pipeline,
        &their_bypass_pipeline,
    )?;

    Ok(())
}

fn collect_chunk_ids(
    pile: &Pile<Arc<Box<dyn RawPile>>>,
    root: &TreeNode,
    chunk_ids: &mut BTreeSet<ChunkId>,
) -> anyhow::Result<()> {
    root.visit(
        &mut |tree_node, _| {
            match tree_node {
                TreeNode::NormalFile { content, .. } => {
                    collect_chunk_ids_from_chunkref(pile, content, chunk_ids)?;
                }
                _ => {}
            }
            Ok(())
        },
        "".to_owned(),
    )?;
    Ok(())
}

fn collect_chunk_ids_from_chunkref(
    pile: &Pile<Arc<Box<dyn RawPile>>>,
    chunk_ref: &RecursiveChunkRef,
    collection: &mut BTreeSet<ChunkId>,
) -> anyhow::Result<()> {
    if chunk_ref.depth == 0 {
        collection.insert(chunk_ref.chunk_id);
    } else {
        let shallower_chunk_ref = RecursiveChunkRef {
            chunk_id: chunk_ref.chunk_id,
            depth: chunk_ref.depth - 1,
        };
        let mut unchunker = RecursiveUnchunker::new(pile, shallower_chunk_ref);
        let mut next_chunk_id: ChunkId = Default::default();
        loop {
            let read = unchunker.read(&mut next_chunk_id[..])?;
            if read == 0 {
                break;
            } else if read < next_chunk_id.len() {
                unchunker.read_exact(&mut next_chunk_id[read..])?;
            }
            collection.insert(next_chunk_id);
        }
    }
    Ok(())
}

pub fn offering_side<R: Read, W: Write>(
    pile: &Pile<Arc<Box<dyn RawPile>>>,
    bypass_pile: &Box<dyn RawPile>,
    reader: &mut R,
    writer: &mut W,
    mut progress: Box<dyn ProgressTracker>,
) -> anyhow::Result<()> {
    let version = env!("CARGO_PKG_VERSION");
    let expecting = format!("Datman Pull Accepter {}", version);
    write_message(writer, &format!("Datman Pull Offerer {}", version))?;
    writer.flush()?;

    let found: String = read_message(reader)?;
    ensure!(
        found == expecting,
        "Version mismatch. Expecting {:?} got {:?}",
        expecting,
        found
    );

    // First 'negotiate' (for now: assert) a pile bypass.
    // This lets us avoid decompressing things before recompressing them at the other end,
    // assuming both ends use the same dictionary.
    negotiate_bypassed_pile(pile, &bypass_pile, reader, writer)?;

    let offered_pointers = offer_pointers(pile, writer)?;
    let wanted_pointers = read_message::<_, BTreeSet<String>>(reader)?;

    let mut chunks_to_offer: BTreeSet<ChunkId> = BTreeSet::new();

    for pointer_name in &wanted_pointers {
        let pointer_data = offered_pointers
            .get(pointer_name)
            .with_context(|| format!("Requested pointer {:?} was not offered", pointer_name))?;

        collect_chunk_ids_from_chunkref(pile, &pointer_data.chunk_ref, &mut chunks_to_offer)?;

        let root_node = retrieve_tree_node(pile, pointer_data.chunk_ref.clone())?;
        collect_chunk_ids(pile, &root_node.node, &mut chunks_to_offer)?;
    }

    write_message(writer, &chunks_to_offer)?;
    writer.flush()?;

    let chunks_to_skip: BTreeSet<ChunkId> = read_message(reader)?;
    let chunks_to_send: Vec<ChunkId> = chunks_to_offer
        .difference(&chunks_to_skip)
        .cloned()
        .collect();

    drop(chunks_to_offer);
    drop(chunks_to_skip);

    progress.set_max_size(chunks_to_send.len() as u64);
    progress.set_current(0);
    for chunk_id in chunks_to_send {
        let chunk_data = bypass_pile
            .read(Keyspace::Chunk, &chunk_id)?
            .context("Chunk vanished")?;

        write_message(writer, &Some((chunk_id, chunk_data)))?;
        progress.inc_progress(1);
    }

    write_message(writer, &None::<Option<(ChunkId, Vec<u8>)>>)?;

    writer.flush()?;

    Ok(())
}

pub fn accepting_side<R: Read, W: Write>(
    pile: &Pile<Arc<Box<dyn RawPile>>>,
    bypass_pile: &Box<dyn RawPile>,
    reader: &mut R,
    writer: &mut W,
    mut progress: Box<dyn ProgressTracker>,
) -> anyhow::Result<()> {
    let version = env!("CARGO_PKG_VERSION");
    let expecting = format!("Datman Pull Offerer {}", version);
    write_message(writer, &format!("Datman Pull Accepter {}", version))?;
    writer.flush()?;

    let found: String = read_message(reader)?;
    ensure!(
        found == expecting,
        "Version mismatch. Expecting {:?} got {:?}",
        expecting,
        found
    );

    // First 'negotiate' (for now: assert) a pile bypass.
    // This lets us avoid decompressing things before recompressing them at the other end,
    // assuming both ends use the same dictionary.
    negotiate_bypassed_pile(pile, &bypass_pile, reader, writer)?;

    let offered_pointers: BTreeMap<String, PointerData> = read_message(reader)?;
    let mut wanted_pointers: BTreeSet<String> = BTreeSet::new();

    for (pointer_name, pointer_data) in &offered_pointers {
        if pile.read_pointer(pointer_name)?.is_none() {
            wanted_pointers.insert(pointer_name.clone());
            if let Some(parent) = &pointer_data.parent_pointer {
                if pile.read_pointer(parent)?.is_none() && !offered_pointers.contains_key(parent) {
                    bail!("Offered pointer {:?} requires parent {:?} which we don't have and isn't offered.", pointer_name, parent);
                }
            }
        }
    }

    write_message(writer, &wanted_pointers)?;
    writer.flush()?;

    let offered_chunks: BTreeSet<ChunkId> = read_message(reader)?;
    let mut chunks_to_skip: BTreeSet<ChunkId> = BTreeSet::new();
    for chunk_id in &offered_chunks {
        if pile.chunk_exists(chunk_id)? {
            chunks_to_skip.insert(*chunk_id);
        }
    }

    write_message(writer, &chunks_to_skip)?;
    writer.flush()?;

    let num_chunks_to_recv = offered_chunks.len() - chunks_to_skip.len();

    let mut chunks_to_recv: BTreeSet<ChunkId> = offered_chunks
        .difference(&chunks_to_skip)
        .cloned()
        .collect();

    drop(offered_chunks);
    drop(chunks_to_skip);

    progress.set_max_size(num_chunks_to_recv as u64);
    progress.set_current(0);

    while let Some((chunk_id, chunk_data)) = read_message::<_, Option<(ChunkId, Vec<u8>)>>(reader)?
    {
        ensure!(
            chunks_to_recv.remove(&chunk_id),
            "Received unexpected chunk"
        );
        bypass_pile.write(Keyspace::Chunk, &chunk_id, &chunk_data)?;
        progress.inc_progress(1);
    }

    ensure!(chunks_to_recv.is_empty(), "Unreceived chunks.");

    for (pointer_name, pointer_data) in &offered_pointers {
        pile.write_pointer(pointer_name, pointer_data)?;
    }
    pile.flush()?;

    Ok(())
}
